﻿Option Strict On
Option Explicit On

Imports System.Runtime.InteropServices

Public Class FileOperation
    Implements IFileOperationProgressSink

#Region "API"
    <DllImport("Shell32.dll", EntryPoint:="SHGetDesktopFolder")>
    Private Shared Function SHGetDesktopFolder(<Out, MarshalAs(UnmanagedType.Interface)> ByRef ppshf As IShellFolder) As Integer
    End Function

    <DllImport("Shell32.dll", EntryPoint:="SHCreateShellItem")>
    Private Shared Function SHCreateShellItem(<[In]> pidlParent As IntPtr,
                                              <[In]> psfParent As IntPtr,
                                              <[In]> pidl As IntPtr,
                                              <Out, MarshalAs(UnmanagedType.Interface)> ByRef ppv As IShellItem) As Integer
    End Function

    <DllImport("Shell32.dll", EntryPoint:="SHCreateShellItemArrayFromIDLists")>
    Private Shared Function SHCreateShellItemArrayFromIDLists(<[In]> cidl As Integer,
                                                              <[In]> rgpidl As IntPtr,
                                                              <Out, MarshalAs(UnmanagedType.Interface)> ByRef ppsiItemArray As IShellItemArray) As Integer
    End Function
#End Region

#Region "Enum"
    Private Enum PDMODE As Integer
        PDM_DEFAULT = &H0
        PDM_RUN = &H1
        PDM_PREFLIGHT = &H2
        PDM_UNDOING = &H4
        PDM_ERRORSBLOCKING = &H8
        PDM_INDETERMINATE = &H10
    End Enum

    Private Enum PDOPSTATUS As Integer
        PDOPS_NOTINITIALIZED = 0
        PDOPS_RUNNING = 1
        PDOPS_PAUSED = 2
        PDOPS_CANCELLED = 3
        PDOPS_STOPPED = 4
        PDOPS_ERRORS = 5
    End Enum

    Private Enum OPPROGDLGF
        OPPROGDLG_DEFAULT = &H0
        OPPROGDLG_MODAL = &H1
        OPPROGDLG_AUTOTIME = &H2
        OPPROGDLG_NOTIME = &H4
        OPPROGDLG_NOMINIMIZE = &H8
        OPPROGDLG_NOPROGRESSBAR = &H10

        'This flag is invalid in this method. To set the progress bar to
        'marquee mode, use the flags in IOperationsProgressDialog::SetMode.
        'OPPROGDLG_MARQUEEPROGRESS = &H20

        OPPROGDLG_NOCANCEL = &H40
        OPPROGDLG_ENABLEPAUSE = &H80
        OPPROGDLG_ALLOWUNDO = &H100
        OPPROGDLG_DONTDISPLAYSOURCEPATH = &H200
        OPPROGDLG_DONTDISPLAYDESTPATH = &H400
        OPPROGDLG_NOMULTIDAYESTIMATES = &H800
        OPPROGDLG_DONTDISPLAYLOCATIONS = &H1000
    End Enum

    Private Enum SPACTION As Integer
        SPACTION_NONE = 0
        SPACTION_MOVING = 1
        SPACTION_COPYING = 2
        SPACTION_RECYCLING = 3
        SPACTION_APPLYINGATTRIBS = 4
        SPACTION_DOWNLOADING = 5
        SPACTION_SEARCHING_INTERNET = 6
        SPACTION_CALCULATING = 7
        SPACTION_UPLOADING = 8
        SPACTION_SEARCHING_FILES = 9
        SPACTION_DELETING = 10
        SPACTION_RENAMING = 11
        SPACTION_FORMATTING = 12
        SPACTION_COPY_MOVING = 13
    End Enum

    Public Enum SIGDN As Integer
        SIGDN_NORMALDISPLAY = &H0
        SIGDN_PARENTRELATIVEPARSING = &H80018001
        SIGDN_DESKTOPABSOLUTEPARSING = &H80028000
        SIGDN_PARENTRELATIVEEDITING = &H80031001
        SIGDN_DESKTOPABSOLUTEEDITING = &H8004C000
        SIGDN_FILESYSPATH = &H80058000
        SIGDN_URL = &H80068000
        SIGDN_PARENTRELATIVEFORADDRESSBAR = &H8007C001
        SIGDN_PARENTRELATIVE = &H80080001
        SIGDN_PARENTRELATIVEFORUI = &H80094001
    End Enum

    Public Enum SFGAO As Integer
        SFGAO_CANCOPY = &H1
        SFGAO_CANMOVE = &H2
        SFGAO_CANLINK = &H4
        SFGAO_STORAGE = &H8
        SFGAO_CANRENAME = &H10
        SFGAO_CANDELETE = &H20
        SFGAO_HASPROPSHEET = &H40
        SFGAO_DROPTARGET = &H100
        SFGAO_CAPABILITYMASK = &H177
        SFGAO_ENCRYPTED = &H2000
        SFGAO_ISSLOW = &H4000
        SFGAO_GHOSTED = &H8000
        SFGAO_LINK = &H10000
        SFGAO_SHARE = &H20000
        SFGAO_READONLY = &H40000
        SFGAO_HIDDEN = &H80000
        SFGAO_DISPLAYATTRMASK = &HFC000
        SFGAO_FILESYSANCESTOR = &H10000000
        SFGAO_FOLDER = &H20000000
        SFGAO_FILESYSTEM = &H40000000
        SFGAO_HASSUBFOLDER = &H80000000
        SFGAO_CONTENTSMASK = &H80000000
        SFGAO_VALIDATE = &H1000000
        SFGAO_REMOVABLE = &H2000000
        SFGAO_COMPRESSED = &H4000000
        SFGAO_BROWSABLE = &H8000000
        SFGAO_NONENUMERATED = &H100000
        SFGAO_NEWCONTENT = &H200000
        SFGAO_CANMONIKER = &H400000
        SFGAO_HASSTORAGE = &H400000
        SFGAO_STREAM = &H400000
        SFGAO_STORAGEANCESTOR = &H800000
        SFGAO_STORAGECAPMASK = &H70C50008
    End Enum

    Public Enum SICHINT As Integer
        SICHINT_DISPLAY = &H0
        SICHINT_ALLFIELDS = &H80000000
        SICHINT_CANONICAL = &H10000000
        SICHINT_TEST_FILESYSPATH_IF_NOT_EQUAL = &H20000000
    End Enum

    Public Enum FILEOPERATIONFLAGS As Integer
        FOF_NOTHING = &H0
        'FOF_MULTIDESTFILES = &H1
        'FOF_CONFIRMMOUSE = &H2
        FOF_SILENT = &H4
        FOF_RENAMEONCOLLISION = &H8
        FOF_NOCONFIRMATION = &H10
        'FOF_WANTMAPPINGHANDLE = &H20
        FOF_ALLOWUNDO = &H40
        FOF_FILESONLY = &H80
        'FOF_SIMPLEPROGRESS = &H100
        FOF_NOCONFIRMMKDIR = &H200
        FOF_NOERRORUI = &H400
        FOF_NORECURSION = &H1000
        FOF_NOCOPYSECURITYATTRIBS = &H800
        FOF_NO_CONNECTED_ELEMENTS = &H2000
        FOF_WANTNUKEWARNING = &H4000
        FOF_NORECURSEREPARSE = &H8000&
        FOF_NO_UI = FOF_SILENT Or FOF_NOCONFIRMATION Or FOF_NOERRORUI Or FOF_NOCONFIRMMKDIR
        FOF_DEFAULT = FOF_ALLOWUNDO Or FOF_NOCONFIRMMKDIR
        FOFX_NOSKIPJUNCTIONS = &H10000
        FOFX_PREFERHARDLINK = &H20000
        FOFX_SHOWELEVATIONPROMPT = &H40000
        FOFX_RECYCLEONDELETE = &H80000
        FOFX_EARLYFAILURE = &H100000
        FOFX_PRESERVEFILEEXTENSIONS = &H200000
        FOFX_KEEPNEWERFILE = &H400000
        FOFX_NOCOPYHOOKS = &H800000
        FOFX_NOMINIMIZEBOX = &H1000000
        FOFX_MOVEACLSACROSSVOLUMES = &H2000000
        FOFX_DONTDISPLAYSOURCEPATH = &H4000000
        FOFX_DONTDISPLAYDESTPATH = &H8000000
        FOFX_REQUIREELEVATION = &H10000000
        FOFX_ADDUNDORECORD = &H20000000
        FOFX_COPYASDOWNLOAD = &H40000000
        FOFX_DONTDISPLAYLOCATIONS = &H80000000
    End Enum

    Public Enum TRANSFER_SOURCE_FLAGS
        TSF_NORMAL = &H0
        TSF_FAIL_EXIST = &H0
        TSF_RENAME_EXIST = &H1
        TSF_OVERWRITE_EXIST = &H2
        TSF_ALLOW_DECRYPTION = &H4
        TSF_NO_SECURITY = &H8
        TSF_COPY_CREATION_TIME = &H10
        TSF_COPY_WRITE_TIME = &H20
        TSF_USE_FULL_ACCESS = &H40
        TSF_DELETE_RECYCLE_IF_POSSIBLE = &H80
        TSF_COPY_HARD_LINK = &H100
        TSF_COPY_LOCALIZED_NAME = &H200
        TSF_MOVE_AS_COPY_DELETE = &H400
        TSF_SUSPEND_SHELLEVENTS = &H800
    End Enum

    Public Enum FILE_ATTRIBUTE As Integer
        FILE_ATTRIBUTE_READONLY = &H1
        FILE_ATTRIBUTE_HIDDEN = &H2
        FILE_ATTRIBUTE_SYSTEM = &H4
        FILE_ATTRIBUTE_DIRECTORY = &H10
        FILE_ATTRIBUTE_ARCHIVE = &H20
        FILE_ATTRIBUTE_NORMAL = &H80
        FILE_ATTRIBUTE_TEMPORARY = &H100
        FILE_ATTRIBUTE_SPARSE_FILE = &H200
        FILE_ATTRIBUTE_REPARSE_POINT = &H400
        FILE_ATTRIBUTE_COMPRESSED = &H800
        FILE_ATTRIBUTE_OFFLINE = &H1000
        FILE_ATTRIBUTE_NOT_CONTENT_INDEXED = &H2000
        FILE_ATTRIBUTE_ENCRYPTED = &H4000
        FILE_ATTRIBUTE_INTEGRITY_STREAM = &H8000
        FILE_ATTRIBUTE_NO_SCRUB_DATA = &H20000
        FILE_ATTRIBUTE_RECALL_ON_OPEN = &H40000
        FILE_ATTRIBUTE_RECALL_ON_DATA_ACCESS = &H400000
    End Enum

    Private Enum VARTYPE As UShort
        VT_EMPTY = 0
        VT_NULL = 1
        VT_I2 = 2
        VT_I4 = 3
        VT_R4 = 4
        VT_R8 = 5
        VT_CY = 6
        VT_DATE = 7
        VT_BSTR = 8
        VT_DISPATCH = 9
        VT_ERROR = 10
        VT_BOOL = 11
        VT_VARIANT = 12
        VT_UNKNOWN = 13
        VT_DECIMAL = 14
        VT_I1 = 16
        VT_UI1 = 17
        VT_UI2 = 18
        VT_UI4 = 19
        VT_I8 = 20
        VT_UI8 = 21
        VT_INT = 22
        VT_UINT = 23
        VT_VOID = 24
        VT_HRESULT = 25
        VT_PTR = 26
        VT_SAFEARRAY = 27
        VT_CARRAY = 28
        VT_USERDEFINED = 29
        VT_LPSTR = 30
        VT_LPWSTR = 31
        VT_RECORD = 36
        VT_FILETIME = 64
        VT_BLOB = 65
        VT_STREAM = 66
        VT_STORAGE = 67
        VT_STREAMED_OBJECT = 68
        VT_STORED_OBJECT = 69
        VT_BLOB_OBJECT = 70
        VT_CF = 71
        VT_CLSID = 72
        VT_ILLEGALMASKED = 4095
        VT_VECTOR = 4096
        VT_ARRAY = 8192
        VT_BYREF = 16384
    End Enum
#End Region

#Region "Const"
    Public Const S_OK As Integer = &H0

    Private Const IID_IShellItem As String = "43826d1e-e718-42ee-bc55-a1e261c37bfe"
    Private Const IID_IShellFolder As String = "000214E6-0000-0000-C000-000000000046"
    Private Const IID_IFileOperation As String = "947aab5f-0a5c-4c13-b4d6-4bf7836fc9f8"
    Private Const IID_IShellItemArray As String = "b63ea76d-1f85-456f-a19c-48159efa858b"
    Private Const IID_IPropertyChange As String = "f917bc8a-1bba-4478-a245-1bde03eb9431"
    Private Const IID_IPropertyChangeArray As String = "380f5cad-1b5e-42f2-805d-637fd392d31e"
    Private Const IID_IOperationsProgressDialog As String = "0c9fb851-e5c9-43eb-a370-f0677b13874c"
    Private Const IID_IFileOperationProgressSink As String = "04b0f1a7-9490-44bc-96e1-4296a31252e2"

    Private Const CLSID_FileOperation As String = "3ad05575-8857-4850-9277-11b85bdb8e09"
    Private Const CLSID_ProgressDialog As String = "f8383852-fcd3-11d1-a6b9-006097df5bd4"
#End Region

#Region "Events"
    Public Event OperationsStart()
    Public Event OperationsFinish(Result As Integer)
    Public Event OperationsUpdateProgress(WorkTotal As UInteger,
                                          WorkSoFar As UInteger)

    Public Event OperationsPreCopyItem(Flags As TRANSFER_SOURCE_FLAGS,
                                       pItem As IntPtr,
                                       pDestinationFolder As IntPtr,
                                       NewName As String)

    Public Event OperationsPostCopyItem(Flags As TRANSFER_SOURCE_FLAGS,
                                        pItem As IntPtr,
                                        pDestinationFolder As IntPtr,
                                        NewName As String,
                                        CopyResult As Integer,
                                        pNewlyCreated As IntPtr)

    Public Event OperationsPreRenameItem(Flags As TRANSFER_SOURCE_FLAGS,
                                         pItem As IntPtr,
                                         pNewName As String)

    Public Event OperationsPostRenameItem(Flags As TRANSFER_SOURCE_FLAGS,
                                          pItem As IntPtr,
                                          NewName As String,
                                          RenameResult As Integer,
                                          pNewlyCreated As IntPtr)

    Public Event OperationsPreMoveItem(Flags As TRANSFER_SOURCE_FLAGS,
                                       pItem As IntPtr,
                                       pDestinationFolder As IntPtr,
                                       NewName As String)

    Public Event OperationsPostMoveItem(Flags As TRANSFER_SOURCE_FLAGS,
                                        pItem As IntPtr,
                                        pDestinationFolder As IntPtr,
                                        pszNewName As String,
                                        MoveResult As Integer,
                                        pNewlyCreated As IntPtr)

    Public Event OperationsPreDeleteItem(Flags As TRANSFER_SOURCE_FLAGS,
                                         pItem As IntPtr)

    Public Event OperationsPostDeleteItem(Flags As TRANSFER_SOURCE_FLAGS,
                                          pItem As IntPtr,
                                          DeleteResult As Integer,
                                          pNewlyCreated As IntPtr)

    Public Event OperationsPreNewItem(Flags As TRANSFER_SOURCE_FLAGS,
                                      pDestinationFolder As IntPtr,
                                      NewName As String)

    Public Event OperationsPostNewItem(Flags As TRANSFER_SOURCE_FLAGS,
                                       pDestinationFolder As IntPtr,
                                       pNewName As String,
                                       TemplateName As String,
                                       FileAttributes As FILE_ATTRIBUTE,
                                       NewResult As Integer,
                                       pNewItem As IntPtr)
#End Region

#Region "Structure"
    Private Structure PROPERTYKEY
        Dim fmtid As Guid
        Dim pid As Integer
    End Structure

    Private Structure CArray
        Dim cElems As UInteger
        Dim pElems As IntPtr
    End Structure

    <StructLayout(LayoutKind.Explicit, Size:=16)>
    Private Structure PROPVARIANT
        <FieldOffset(0)> Dim vt As VARTYPE
        <FieldOffset(2)> Dim wReserved1 As UShort
        <FieldOffset(4)> Dim wReserved2 As UShort
        <FieldOffset(6)> Dim wReserved3 As UShort
        <FieldOffset(8)> Dim uintVal As UInteger 'UINT32
        <FieldOffset(8)> Dim ulVal As ULong      'UINT64
        <FieldOffset(8)> Dim dblVal As Double    'DOUBLE
        <FieldOffset(8)> Dim pwszVal As IntPtr   'STRING
        <FieldOffset(8)> Dim punkVal As IntPtr   'IUNKNOWN
        <FieldOffset(8)> Dim cArray As CArray    'BLOB
        ' ... usw.
    End Structure
#End Region

#Region "Variablen"
    Private m_Cookie As Integer = 0
    Private m_Owner As IWin32Window = Nothing
    Private m_OperationsAborted As Boolean = False
    Private m_IFileOperation As IFileOperation = Nothing
    Private m_UseOldProgressDialog As Boolean = False
    Private m_IOperationsProgressDialog As IOperationsProgressDialog = Nothing
    Private m_FILEOPERATIONFLAGS As FILEOPERATIONFLAGS = FILEOPERATIONFLAGS.FOF_DEFAULT
#End Region

#Region "Properties"
    Public WriteOnly Property Owner As IWin32Window
        Set(value As IWin32Window)
            m_Owner = value
        End Set
    End Property

    Public ReadOnly Property GetOperationsAborted As Boolean
        Get
            Return m_OperationsAborted
        End Get
    End Property

    Public Property OperationFlags As FILEOPERATIONFLAGS
        Get
            Return m_FILEOPERATIONFLAGS
        End Get
        Set(value As FILEOPERATIONFLAGS)
            m_FILEOPERATIONFLAGS = value
        End Set
    End Property

    Public Property UseOldProgressDialog As Boolean
        Get
            Return m_UseOldProgressDialog
        End Get
        Set(value As Boolean)
            m_UseOldProgressDialog = value
        End Set
    End Property
#End Region

#Region "Class Finalize"
    Protected Overrides Sub Finalize()
        DestroyIFileOperation()
        MyBase.Finalize()
    End Sub
#End Region

#Region "Public Functions"
    Public Function PerformOperations() As Boolean
        Dim bolRet As Boolean = False
        If m_IFileOperation.PerformOperations() = S_OK Then
            bolRet = True
        End If
        DestroyIFileOperation()
        Return bolRet
    End Function

    Public Function CopyItem(SourceFileOrFolder As String,
                             DestinationFolder As String,
                             Optional NewName As String = Nothing) As Boolean

        Dim bolRet As Boolean = False

        Dim ShellItem_Src As IShellItem = GetIShellItemFromFileOrFolder(SourceFileOrFolder)

        If ShellItem_Src IsNot Nothing Then

            Dim ShellItem_Dest As IShellItem = GetIShellItemFromFileOrFolder(DestinationFolder)

            If ShellItem_Dest IsNot Nothing Then

                If CreateIFileOperation() Then

                    If m_IFileOperation.CopyItem(ShellItem_Src, ShellItem_Dest, NewName, Nothing) = S_OK Then

                        bolRet = True

                    End If

                End If

                Marshal.ReleaseComObject(ShellItem_Dest)

            End If

            Marshal.ReleaseComObject(ShellItem_Src)

        End If

        Return bolRet

    End Function

    Public Function CopyItems(SourceFilesOrFolders As List(Of String),
                              DestinationFolder As String) As Boolean

        Dim bolRet As Boolean = False
        Dim pIDLList As New List(Of IntPtr)

        If SourceFilesOrFolders IsNot Nothing AndAlso SourceFilesOrFolders.Count > 0 Then

            For Each SourceFileOrFolder In SourceFilesOrFolders

                Dim pIDL As IntPtr = GetPIDLFromFileOrFolder(SourceFileOrFolder)

                If pIDL <> IntPtr.Zero Then

                    pIDLList.Add(pIDL)

                End If

            Next

            Dim ShellItemArray As IShellItemArray = GetIShellItemArrayFromPIDLList(pIDLList)

            If ShellItemArray IsNot Nothing Then

                Dim ShellItem_Dest As IShellItem = GetIShellItemFromFileOrFolder(DestinationFolder)

                If ShellItem_Dest IsNot Nothing Then

                    If CreateIFileOperation() Then

                        If m_IFileOperation.CopyItems(ShellItemArray, ShellItem_Dest) = S_OK Then

                            bolRet = True

                        End If

                    End If

                    Marshal.ReleaseComObject(ShellItem_Dest)

                End If

                Marshal.ReleaseComObject(ShellItemArray)

            End If

            For Each pIDL In pIDLList

                If pIDL <> IntPtr.Zero Then

                    Marshal.FreeCoTaskMem(pIDL)

                End If

            Next

        End If

        Return bolRet

    End Function

    Public Function DeleteItem(SourceFileOrFolder As String) As Boolean

        Dim bolRet As Boolean = False

        Dim ShellItem_Src As IShellItem = GetIShellItemFromFileOrFolder(SourceFileOrFolder)

        If ShellItem_Src IsNot Nothing Then

            If CreateIFileOperation() Then

                If m_IFileOperation.DeleteItem(ShellItem_Src, Nothing) = S_OK Then

                    bolRet = True

                End If

            End If

            Marshal.ReleaseComObject(ShellItem_Src)

        End If

        Return bolRet

    End Function

    Public Function DeleteItems(SourceFilesOrFolders As List(Of String)) As Boolean

        Dim bolRet As Boolean = False
        Dim pIDLList As New List(Of IntPtr)

        If SourceFilesOrFolders IsNot Nothing AndAlso SourceFilesOrFolders.Count > 0 Then

            For Each SourceFileOrFolder In SourceFilesOrFolders

                Dim pIDL As IntPtr = GetPIDLFromFileOrFolder(SourceFileOrFolder)

                If pIDL <> IntPtr.Zero Then

                    pIDLList.Add(pIDL)

                End If

            Next

            Dim ShellItemArray As IShellItemArray = GetIShellItemArrayFromPIDLList(pIDLList)

            If ShellItemArray IsNot Nothing Then

                If CreateIFileOperation() Then

                    If m_IFileOperation.DeleteItems(ShellItemArray) = S_OK Then

                        bolRet = True

                    End If

                    DestroyIFileOperation()

                End If

                Marshal.ReleaseComObject(ShellItemArray)

            End If

            For Each pIDL In pIDLList

                If pIDL <> IntPtr.Zero Then

                    Marshal.FreeCoTaskMem(pIDL)

                End If

            Next

        End If

        Return bolRet

    End Function

    Public Function RenmeItem(SourceFileOrFolder As String,
                              NewName As String) As Boolean

        Dim bolRet As Boolean = False

        Dim ShellItem_Src As IShellItem = GetIShellItemFromFileOrFolder(SourceFileOrFolder)

        If ShellItem_Src IsNot Nothing Then

            If CreateIFileOperation() Then

                If m_IFileOperation.RenameItem(ShellItem_Src, NewName, Nothing) = S_OK Then

                    bolRet = True

                End If

            End If

            Marshal.ReleaseComObject(ShellItem_Src)

        End If

        Return bolRet

    End Function

    Public Function RenmeItems(SourceFilesOrFolders As List(Of String),
                               NewName As String) As Boolean

        Dim bolRet As Boolean = False
        Dim pIDLList As New List(Of IntPtr)

        If SourceFilesOrFolders IsNot Nothing AndAlso SourceFilesOrFolders.Count > 0 Then

            For Each SourceFileOrFolder In SourceFilesOrFolders

                Dim pIDL As IntPtr = GetPIDLFromFileOrFolder(SourceFileOrFolder)

                If pIDL <> IntPtr.Zero Then

                    pIDLList.Add(pIDL)

                End If

            Next

            Dim ShellItemArray As IShellItemArray = GetIShellItemArrayFromPIDLList(pIDLList)

            If ShellItemArray IsNot Nothing Then

                If CreateIFileOperation() Then

                    If m_IFileOperation.RenameItems(ShellItemArray, NewName) = S_OK Then

                        bolRet = True

                    End If

                End If

                Marshal.ReleaseComObject(ShellItemArray)

            End If

            For Each pIDL In pIDLList

                If pIDL <> IntPtr.Zero Then

                    Marshal.FreeCoTaskMem(pIDL)

                End If

            Next

        End If

        Return bolRet
    End Function

    Public Function MoveItem(SourceFileOrFolder As String,
                             DestinationFolder As String,
                             Optional NewName As String = Nothing) As Boolean

        Dim bolRet As Boolean = False

        Dim ShellItem_Src As IShellItem = GetIShellItemFromFileOrFolder(SourceFileOrFolder)

        If ShellItem_Src IsNot Nothing Then

            Dim ShellItem_Dest As IShellItem = GetIShellItemFromFileOrFolder(DestinationFolder)

            If ShellItem_Dest IsNot Nothing Then

                If CreateIFileOperation() Then

                    If m_IFileOperation.MoveItem(ShellItem_Src, ShellItem_Dest, NewName, Nothing) = S_OK Then

                        bolRet = True

                    End If

                End If

                Marshal.ReleaseComObject(ShellItem_Dest)

            End If

            Marshal.ReleaseComObject(ShellItem_Src)

        End If

        Return bolRet

    End Function

    Public Function MoveItems(SourceFilesOrFolders As List(Of String), DestinationFolder As String) As Boolean

        Dim bolRet As Boolean = False
        Dim pIDLList As New List(Of IntPtr)

        If SourceFilesOrFolders IsNot Nothing AndAlso SourceFilesOrFolders.Count > 0 Then

            For Each SourceFileOrFolder In SourceFilesOrFolders

                Dim pIDL As IntPtr = GetPIDLFromFileOrFolder(SourceFileOrFolder)

                If pIDL <> IntPtr.Zero Then

                    pIDLList.Add(pIDL)

                End If

            Next

            Dim ShellItemArray As IShellItemArray = GetIShellItemArrayFromPIDLList(pIDLList)

            If ShellItemArray IsNot Nothing Then

                Dim ShellItem_Dest As IShellItem = GetIShellItemFromFileOrFolder(DestinationFolder)

                If ShellItem_Dest IsNot Nothing Then

                    If CreateIFileOperation() Then

                        If m_IFileOperation.MoveItems(ShellItemArray, ShellItem_Dest) = S_OK Then

                            bolRet = True

                        End If

                        DestroyIFileOperation()

                    End If

                    Marshal.ReleaseComObject(ShellItem_Dest)

                End If

                Marshal.ReleaseComObject(ShellItemArray)

            End If

            For Each pIDL In pIDLList

                If pIDL <> IntPtr.Zero Then

                    Marshal.FreeCoTaskMem(pIDL)

                End If

            Next

        End If

        Return bolRet

    End Function

    Public Function NewItem(DestinationFolder As String,
                            NewFileOrFolder As String,
                            FileAttributes As FILE_ATTRIBUTE,
                            Optional TemplateName As String = Nothing) As Boolean

        Dim bolRet As Boolean = False

        Dim ShellItem_Dst As IShellItem = GetIShellItemFromFileOrFolder(DestinationFolder)

        If ShellItem_Dst IsNot Nothing Then

            If CreateIFileOperation() Then

                If m_IFileOperation.NewItem(ShellItem_Dst, FileAttributes, NewFileOrFolder, TemplateName, Nothing) = S_OK Then

                    bolRet = True

                End If

            End If

            Marshal.ReleaseComObject(ShellItem_Dst)

        End If

        Return bolRet

    End Function
#End Region

#Region "Private Functions"
    Private Function CreateIFileOperation() As Boolean

        Dim bolRet As Boolean = False

        m_OperationsAborted = False

        If m_IFileOperation Is Nothing Then

            m_IFileOperation = CType(Activator.CreateInstance(Type.GetTypeFromCLSID(New Guid(CLSID_FileOperation))), IFileOperation)

            If m_IFileOperation IsNot Nothing Then

                If m_IFileOperation.SetOwnerWindow(If(m_Owner IsNot Nothing, m_Owner.Handle, Form.ActiveForm.Handle)) = S_OK Then

                    If m_IFileOperation.SetOperationFlags(m_FILEOPERATIONFLAGS) = S_OK Then

                        If m_IFileOperation.Advise(Me, m_Cookie) = S_OK Then

                            bolRet = True

                            If Not CBool(m_FILEOPERATIONFLAGS And FILEOPERATIONFLAGS.FOF_SILENT) Then

                                ' ----==== Optional ====----
                                If m_UseOldProgressDialog Then

                                    bolRet = False

                                    If m_IOperationsProgressDialog Is Nothing Then

                                        m_IOperationsProgressDialog = CType(Activator.CreateInstance(Type.GetTypeFromCLSID(New Guid(CLSID_ProgressDialog))), IOperationsProgressDialog)

                                        If m_IOperationsProgressDialog IsNot Nothing Then

                                            If m_IFileOperation.SetProgressDialog(m_IOperationsProgressDialog) = S_OK Then bolRet = True

                                        End If

                                    End If

                                End If
                                ' ----==================----

                            End If

                        End If

                    End If

                End If

            End If

        Else

            bolRet = True

        End If

        Return bolRet

    End Function

    Private Function DestroyIFileOperation() As Boolean

        Dim bolRet As Boolean = False

        If m_IFileOperation IsNot Nothing Then

            m_IFileOperation.GetAnyOperationsAborted(m_OperationsAborted)

            If m_Cookie <> 0 Then

                If m_IFileOperation.Unadvise(m_Cookie) = S_OK Then m_Cookie = 0

            End If

            If m_IOperationsProgressDialog IsNot Nothing Then

                If Marshal.ReleaseComObject(m_IOperationsProgressDialog) = 0 Then m_IOperationsProgressDialog = Nothing

            End If

            If Marshal.ReleaseComObject(m_IFileOperation) = 0 Then
                m_IFileOperation = Nothing
                bolRet = True
            End If

        End If

        Return bolRet

    End Function

    Private Function GetIShellItemFromFileOrFolder(FileOrFolder As String) As IShellItem

        Dim ShellItem As IShellItem = Nothing

        Dim pIDL As IntPtr = GetPIDLFromFileOrFolder(FileOrFolder)

        If pIDL <> IntPtr.Zero Then

            SHCreateShellItem(IntPtr.Zero, IntPtr.Zero, pIDL, ShellItem)

            Marshal.FreeCoTaskMem(pIDL)

        End If

        Return ShellItem

    End Function

    Private Function GetIShellItemArrayFromPIDLList(pIDLList As List(Of IntPtr)) As IShellItemArray

        Dim ShellItemArray As IShellItemArray = Nothing

        If pIDLList.Count > 0 Then

            Dim pIDLArray As GCHandle = GCHandle.Alloc(pIDLList.ToArray, GCHandleType.Pinned)

            SHCreateShellItemArrayFromIDLists(pIDLList.Count, pIDLArray.AddrOfPinnedObject(), ShellItemArray)

            pIDLArray.Free()

        End If

        Return ShellItemArray

    End Function

    Private Function GetPIDLFromFileOrFolder(FileOrFolder As String) As IntPtr

        Dim pIDL As IntPtr = IntPtr.Zero
        Dim ShellFolder As IShellFolder = Nothing

        If SHGetDesktopFolder(ShellFolder) = S_OK Then

            If ShellFolder IsNot Nothing Then

                ShellFolder.ParseDisplayName(IntPtr.Zero, IntPtr.Zero, FileOrFolder, 0, pIDL, 0)

                Marshal.ReleaseComObject(ShellFolder)

            End If

        End If

        Return pIDL

    End Function
#End Region

#Region "Interface IShellItem"
    <ComImport>
    <InterfaceType(ComInterfaceType.InterfaceIsIUnknown)>
    <Guid(IID_IShellItem)>
    Public Interface IShellItem

        'virtual HRESULT STDMETHODCALLTYPE BindToHandler( 
        '    /* [unique][in] */ __RPC__in_opt IBindCtx *pbc,
        '    /* [in] */ __RPC__in REFGUID bhid,
        '    /* [in] */ __RPC__in REFIID riid,
        '    /* [iid_is][out] */ __RPC__deref_out_opt void **ppv) = 0;
        <PreserveSig> Function BindToHandler(<[In]> pbc As IntPtr,
                                             <[In]> ByRef bhid As Guid,
                                             <[In]> ByRef riid As Guid,
                                             <Out> ppv As IntPtr) As Integer
        <PreserveSig> Function GetParent(<Out, MarshalAs(UnmanagedType.Interface)> ByRef ppsi As IShellItem) As Integer
        <PreserveSig> Function GetDisplayName(<[In]> sigdnName As SIGDN,
                                              <Out, MarshalAs(UnmanagedType.LPWStr)> ByRef ppszName As String) As Integer
        <PreserveSig> Function GetAttributes(<[In]> sfgaoMask As SFGAO,
                                             <Out> ByRef psfgaoAttribs As SFGAO) As Integer
        <PreserveSig> Function Compare(<[In], MarshalAs(UnmanagedType.Interface)> psi As IShellItem,
                                       <[In]> hint As SICHINT,
                                       <Out> ByRef piOrder As Integer) As Integer

    End Interface
#End Region

#Region "Interface IShellFolder"
    <ComImport>
    <InterfaceType(ComInterfaceType.InterfaceIsIUnknown)>
    <Guid(IID_IShellFolder)>
    Private Interface IShellFolder

        'virtual HRESULT STDMETHODCALLTYPE ParseDisplayName( 
        '    /* [unique][in] */ __RPC__in_opt HWND hwnd,
        '    /* [unique][in] */ __RPC__in_opt IBindCtx *pbc,
        '    /* [string][in] */ __RPC__in_string LPWSTR pszDisplayName,
        '    /* [annotation][unique][out][in] */ 
        '    _Reserved_  ULONG *pchEaten,
        '    /* [out] */ __RPC__deref_out_opt PIDLIST_RELATIVE *ppidl,
        '    /* [unique][out][in] */ __RPC__inout_opt ULONG *pdwAttributes) = 0;
        <PreserveSig> Function ParseDisplayName(<[In]> hwnd As IntPtr,
                                                <[In]> pbc As IntPtr,
                                                <[In], MarshalAs(UnmanagedType.LPWStr)> pszDisplayName As String,
                                                <[In], Out> ByRef pchEaten As Integer,
                                                <Out> ByRef ppidl As IntPtr,
                                                <[In], Out> ByRef pdwAttributes As Integer) As Integer

        'virtual HRESULT STDMETHODCALLTYPE EnumObjects( 
        '    /* [unique][in] */ __RPC__in_opt HWND hwnd,
        '    /* [in] */ SHCONTF grfFlags,
        '    /* [out] */ __RPC__deref_out_opt IEnumIDList **ppenumIDList) = 0;

        'virtual HRESULT STDMETHODCALLTYPE BindToObject( 
        '    /* [in] */ __RPC__in PCUIDLIST_RELATIVE pidl,
        '    /* [unique][in] */ __RPC__in_opt IBindCtx *pbc,
        '    /* [in] */ __RPC__in REFIID riid,
        '    /* [iid_is][out] */ __RPC__deref_out_opt void **ppv) = 0;

        'virtual HRESULT STDMETHODCALLTYPE BindToStorage( 
        '    /* [in] */ __RPC__in PCUIDLIST_RELATIVE pidl,
        '    /* [unique][in] */ __RPC__in_opt IBindCtx *pbc,
        '    /* [in] */ __RPC__in REFIID riid,
        '    /* [iid_is][out] */ __RPC__deref_out_opt void **ppv) = 0;

        'virtual HRESULT STDMETHODCALLTYPE CompareIDs( 
        '    /* [in] */ LPARAM lParam,
        '    /* [in] */ __RPC__in PCUIDLIST_RELATIVE pidl1,
        '    /* [in] */ __RPC__in PCUIDLIST_RELATIVE pidl2) = 0;

        'virtual HRESULT STDMETHODCALLTYPE CreateViewObject( 
        '    /* [unique][in] */ __RPC__in_opt HWND hwndOwner,
        '    /* [in] */ __RPC__in REFIID riid,
        '    /* [iid_is][out] */ __RPC__deref_out_opt void **ppv) = 0;

        'virtual HRESULT STDMETHODCALLTYPE GetAttributesOf( 
        '    /* [in] */ UINT cidl,
        '    /* [unique][size_is][in] */ __RPC__in_ecount_full_opt(cidl) PCUITEMID_CHILD_ARRAY apidl,
        '    /* [out][in] */ __RPC__inout SFGAOF *rgfInOut) = 0;

        'virtual HRESULT STDMETHODCALLTYPE GetUIObjectOf( 
        '    /* [unique][in] */ __RPC__in_opt HWND hwndOwner,
        '    /* [in] */ UINT cidl,
        '    /* [unique][size_is][in] */ __RPC__in_ecount_full_opt(cidl) PCUITEMID_CHILD_ARRAY apidl,
        '    /* [in] */ __RPC__in REFIID riid,
        '    /* [annotation][unique][out][in] */ 
        '    _Reserved_  UINT *rgfReserved,
        '    /* [iid_is][out] */ __RPC__deref_out_opt void **ppv) = 0;

        'virtual HRESULT STDMETHODCALLTYPE GetDisplayNameOf( 
        '    /* [unique][in] */ __RPC__in_opt PCUITEMID_CHILD pidl,
        '    /* [in] */ SHGDNF uFlags,
        '    /* [out] */ __RPC__out STRRET *pName) = 0;

        'virtual /* [local] */ HRESULT STDMETHODCALLTYPE SetNameOf( 
        '    /* [annotation][unique][in] */ 
        '    _In_opt_  HWND hwnd,
        '    /* [annotation][in] */ 
        '    _In_  PCUITEMID_CHILD pidl,
        '    /* [annotation][string][in] */ 
        '    _In_  LPCWSTR pszName,
        '    /* [annotation][in] */ 
        '    _In_  SHGDNF uFlags,
        '    /* [annotation][out] */ 
        '    _Outptr_opt_  PITEMID_CHILD *ppidlOut) = 0;

    End Interface
#End Region

#Region "Interface IFileOperation"
    <ComImport>
    <InterfaceType(ComInterfaceType.InterfaceIsIUnknown)>
    <Guid(IID_IFileOperation)>
    Private Interface IFileOperation
        <PreserveSig> Function Advise(<[In], MarshalAs(UnmanagedType.Interface)> pfops As IFileOperationProgressSink,
                                      <Out> ByRef pdwCookie As Integer) As Integer
        <PreserveSig> Function Unadvise(<[In]> dwCookie As Integer) As Integer
        <PreserveSig> Function SetOperationFlags(<[In]> dwOperationFlags As FILEOPERATIONFLAGS) As Integer

        ' Not implemented!
        <PreserveSig> Function SetProgressMessage(<[In], MarshalAs(UnmanagedType.LPWStr)> pszMessage As String) As Integer

        <PreserveSig> Function SetProgressDialog(<[In], MarshalAs(UnmanagedType.Interface)> popd As IOperationsProgressDialog) As Integer
        <PreserveSig> Function SetProperties(<[In], MarshalAs(UnmanagedType.Interface)> pproparray As IPropertyChangeArray) As Integer
        <PreserveSig> Function SetOwnerWindow(<[In]> hwndOwner As IntPtr) As Integer
        <PreserveSig> Function ApplyPropertiesToItem(<[In], MarshalAs(UnmanagedType.Interface)> psiItem As IShellItem) As Integer
        <PreserveSig> Function ApplyPropertiesToItems(<[In], MarshalAs(UnmanagedType.Interface)> punkItems As IShellItemArray) As Integer
        <PreserveSig> Function RenameItem(<[In], MarshalAs(UnmanagedType.Interface)> psiItem As IShellItem,
                                          <[In], MarshalAs(UnmanagedType.LPWStr)> pszNewName As String,
                                          <[In], MarshalAs(UnmanagedType.Interface)> pfopsItem As IFileOperationProgressSink) As Integer
        <PreserveSig> Function RenameItems(<[In], MarshalAs(UnmanagedType.Interface)> punkItems As IShellItemArray,
                                           <[In], MarshalAs(UnmanagedType.LPWStr)> pszNewName As String) As Integer
        <PreserveSig> Function MoveItem(<[In], MarshalAs(UnmanagedType.Interface)> psiItem As IShellItem,
                                        <[In], MarshalAs(UnmanagedType.Interface)> psiDestinationFolder As IShellItem,
                                        <[In], MarshalAs(UnmanagedType.LPWStr)> pszNewName As String,
                                        <[In], MarshalAs(UnmanagedType.Interface)> pfopsItem As IFileOperationProgressSink) As Integer
        <PreserveSig> Function MoveItems(<[In], MarshalAs(UnmanagedType.Interface)> punkItems As IShellItemArray,
                                         <[In], MarshalAs(UnmanagedType.Interface)> psiDestinationFolder As IShellItem) As Integer
        <PreserveSig> Function CopyItem(<[In], MarshalAs(UnmanagedType.Interface)> psiItem As IShellItem,
                                        <[In], MarshalAs(UnmanagedType.Interface)> psiDestinationFolder As IShellItem,
                                        <[In], MarshalAs(UnmanagedType.LPWStr)> pszCopyName As String,
                                        <[In], MarshalAs(UnmanagedType.Interface)> pfopsItem As IFileOperationProgressSink) As Integer

        <PreserveSig> Function CopyItems(<[In], MarshalAs(UnmanagedType.Interface)> punkItems As IShellItemArray,
                                         <[In], MarshalAs(UnmanagedType.Interface)> psiDestinationFolder As IShellItem) As Integer
        <PreserveSig> Function DeleteItem(<[In], MarshalAs(UnmanagedType.Interface)> psiItem As IShellItem,
                                          <[In], MarshalAs(UnmanagedType.Interface)> pfopsItem As IFileOperationProgressSink) As Integer
        <PreserveSig> Function DeleteItems(<[In], MarshalAs(UnmanagedType.Interface)> punkItems As IShellItemArray) As Integer
        <PreserveSig> Function NewItem(<[In], MarshalAs(UnmanagedType.Interface)> psiDestinationFolder As IShellItem,
                                       <[In]> dwFileAttributes As FILE_ATTRIBUTE,
                                       <[In], MarshalAs(UnmanagedType.LPWStr)> pszName As String,
                                       <[In], MarshalAs(UnmanagedType.LPWStr)> pszTemplateName As String,
                                       <[In], MarshalAs(UnmanagedType.Interface)> pfopsItem As IFileOperationProgressSink) As Integer
        <PreserveSig> Function PerformOperations() As Integer
        <PreserveSig> Function GetAnyOperationsAborted(<Out, MarshalAs(UnmanagedType.Bool)> ByRef pfAnyOperationsAborted As Boolean) As Integer

    End Interface
#End Region

#Region "Interface IShellItemArray"
    <ComImport>
    <InterfaceType(ComInterfaceType.InterfaceIsIUnknown)>
    <Guid(IID_IShellItemArray)>
    Private Interface IShellItemArray

        'virtual HRESULT STDMETHODCALLTYPE BindToHandler( 
        '    /* [unique][in] */ __RPC__in_opt IBindCtx *pbc,
        '    /* [in] */ __RPC__in REFGUID bhid,
        '    /* [in] */ __RPC__in REFIID riid,
        '    /* [iid_is][out] */ __RPC__deref_out_opt void **ppvOut) = 0;
        <PreserveSig> Function BindToHandler() As Integer

        'virtual HRESULT STDMETHODCALLTYPE GetPropertyStore( 
        '    /* [in] */ GETPROPERTYSTOREFLAGS flags,
        '    /* [in] */ __RPC__in REFIID riid,
        '    /* [iid_is][out] */ __RPC__deref_out_opt void **ppv) = 0;
        <PreserveSig> Function GetPropertyStore() As Integer

        'virtual HRESULT STDMETHODCALLTYPE GetPropertyDescriptionList( 
        '    /* [in] */ __RPC__in REFPROPERTYKEY keyType,
        '    /* [in] */ __RPC__in REFIID riid,
        '    /* [iid_is][out] */ __RPC__deref_out_opt void **ppv) = 0;
        <PreserveSig> Function GetPropertyDescriptionList() As Integer

        'virtual HRESULT STDMETHODCALLTYPE GetAttributes( 
        '    /* [in] */ SIATTRIBFLAGS AttribFlags,
        '    /* [in] */ SFGAOF sfgaoMask,
        '    /* [out] */ __RPC__out SFGAOF *psfgaoAttribs) = 0;
        <PreserveSig> Function GetAttributes() As Integer

        'virtual HRESULT STDMETHODCALLTYPE GetCount( 
        '    /* [out] */ __RPC__out DWORD *pdwNumItems) = 0;
        <PreserveSig> Function GetCount() As Integer

        'virtual HRESULT STDMETHODCALLTYPE GetItemAt( 
        '    /* [in] */ DWORD dwIndex,
        '    /* [out] */ __RPC__deref_out_opt IShellItem **ppsi) = 0;
        <PreserveSig> Function GetItemAt() As Integer

        'virtual HRESULT STDMETHODCALLTYPE EnumItems( 
        '    /* [out] */ __RPC__deref_out_opt IEnumShellItems **ppenumShellItems) = 0
        <PreserveSig> Function EnumItems() As Integer

    End Interface
#End Region

#Region "Interface IPropertyChange"
    <ComImport>
    <InterfaceType(ComInterfaceType.InterfaceIsIUnknown)>
    <Guid(IID_IPropertyChange)>
    Private Interface IPropertyChange

        ' ----==== IObjectWithPropertyKey ====----
        'virtual HRESULT STDMETHODCALLTYPE SetPropertyKey( 
        '    /* [in] */ __RPC__in REFPROPERTYKEY key) = 0;
        <PreserveSig> Function SetPropertyKey(<[In]> ByRef key As PROPERTYKEY) As Integer

        'virtual HRESULT STDMETHODCALLTYPE GetPropertyKey( 
        '    /* [out] */ __RPC__out PROPERTYKEY *pkey) = 0;
        <PreserveSig> Function GetPropertyKey(<Out> ByRef pkey As PROPERTYKEY) As Integer

        ' ----==== IPropertyChange ====----
        'virtual HRESULT STDMETHODCALLTYPE ApplyToPropVariant( 
        '    /* [in] */ __RPC__in REFPROPVARIANT propvarIn,
        '    /* [out] */ __RPC__out PROPVARIANT *ppropvarOut) = 0;
        <PreserveSig> Function ApplyToPropVariant(<[In]> ByRef propvarIn As PROPVARIANT,
                                                  <Out> ByRef ppropvarOut As PROPVARIANT) As Integer

    End Interface
#End Region

#Region "Interface IPropertyChangeArray"
    <ComImport>
    <InterfaceType(ComInterfaceType.InterfaceIsIUnknown)>
    <Guid(IID_IPropertyChangeArray)>
    Private Interface IPropertyChangeArray

        <PreserveSig> Function GetCount(<Out> ByRef pcOperations As UInteger) As Integer
        <PreserveSig> Function GetAt(<[In]> iIndex As UInteger,
                                     <[In]> ByRef riid As Guid,
                                     <Out, MarshalAs(UnmanagedType.Interface)> ppv As Object) As Integer
        <PreserveSig> Function InsertAt(<[In]> iIndex As UInteger,
                                        <[In], MarshalAs(UnmanagedType.Interface)> ppropChange As IPropertyChange) As Integer
        <PreserveSig> Function Append(<[In], MarshalAs(UnmanagedType.Interface)> ppropChange As IPropertyChange) As Integer
        <PreserveSig> Function AppendOrReplace(<[In], MarshalAs(UnmanagedType.Interface)> ppropChange As IPropertyChange) As Integer
        <PreserveSig> Function RemoveAt(<[In]> iIndex As UInteger) As Integer
        <PreserveSig> Function IsKeyInArray(<[In]> ByRef key As PROPERTYKEY) As Integer

    End Interface
#End Region

#Region "Interface IOperationsProgressDialog"
    <ComImport>
    <InterfaceType(ComInterfaceType.InterfaceIsIUnknown)>
    <Guid(IID_IOperationsProgressDialog)>
    Private Interface IOperationsProgressDialog

        <PreserveSig> Function StartProgressDialog(<[In]> hwndOwner As IntPtr,
                                                   <[In]> flags As OPPROGDLGF) As Integer
        <PreserveSig> Function StopProgressDialog() As Integer
        <PreserveSig> Function SetOperation(<[In]> action As SPACTION) As Integer
        <PreserveSig> Function SetMode(<[In]> mode As PDMODE) As Integer
        <PreserveSig> Function UpdateProgress(<[In]> ullPointsCurrent As ULong,
                                              <[In]> ullPointsTotal As ULong,
                                              <[In]> ullSizeCurrent As ULong,
                                              <[In]> ullSizeTotal As ULong,
                                              <[In]> ullItemsCurrent As ULong,
                                              <[In]> ullItemsTotal As ULong) As Integer
        <PreserveSig> Function UpdateLocations(<[In], MarshalAs(UnmanagedType.Interface)> psiSource As IShellItem,
                                               <[In], MarshalAs(UnmanagedType.Interface)> psiTarget As IShellItem,
                                               <[In], MarshalAs(UnmanagedType.Interface)> psiItem As IShellItem) As Integer

        <PreserveSig> Function ResetTimer() As Integer
        <PreserveSig> Function PauseTimer() As Integer
        <PreserveSig> Function ResumeTimer() As Integer
        <PreserveSig> Function GetMilliseconds(<Out> ByRef pullElapsed As ULong,
                                               <Out> ByRef pullRemaining As ULong) As Integer
        <PreserveSig> Function GetOperationStatus(<Out> ByRef popstatus As PDOPSTATUS) As Integer

    End Interface
#End Region

#Region "Interface IFileOperationProgressSink"
    <ComImport>
    <InterfaceType(ComInterfaceType.InterfaceIsIUnknown)>
    <Guid(IID_IFileOperationProgressSink)>
    Private Interface IFileOperationProgressSink
        <PreserveSig> Function StartOperations() As Integer
        <PreserveSig> Function FinishOperations(<[In]> hrResult As Integer) As Integer
        <PreserveSig> Function PreRenameItem(<[In]> dwFlags As TRANSFER_SOURCE_FLAGS,
                                             <[In]> psiItem As IntPtr,
                                             <[In], MarshalAs(UnmanagedType.LPWStr)> pszNewName As String) As Integer
        <PreserveSig> Function PostRenameItem(<[In]> dwFlags As TRANSFER_SOURCE_FLAGS,
                                              <[In]> psiItem As IntPtr,
                                              <[In], MarshalAs(UnmanagedType.LPWStr)> pszNewName As String,
                                              <[In]> hrRename As Integer,
                                              <[In]> psiNewlyCreated As IntPtr) As Integer
        <PreserveSig> Function PreMoveItem(<[In]> dwFlags As TRANSFER_SOURCE_FLAGS,
                                           <[In]> psiItem As IntPtr,
                                           <[In]> psiDestinationFolder As IntPtr,
                                           <[In], MarshalAs(UnmanagedType.LPWStr)> pszNewName As String) As Integer
        <PreserveSig> Function PostMoveItem(<[In]> dwFlags As TRANSFER_SOURCE_FLAGS,
                                            <[In]> psiItem As IntPtr,
                                            <[In]> psiDestinationFolder As IntPtr,
                                            <[In], MarshalAs(UnmanagedType.LPWStr)> pszNewName As String,
                                            <[In]> hrMove As Integer,
                                            <[In]> psiNewlyCreated As IntPtr) As Integer
        <PreserveSig> Function PreCopyItem(<[In]> dwFlags As TRANSFER_SOURCE_FLAGS,
                                           <[In]> psiItem As IntPtr,
                                           <[In]> psiDestinationFolder As IntPtr,
                                           <[In], MarshalAs(UnmanagedType.LPWStr)> pszNewName As String) As Integer
        <PreserveSig> Function PostCopyItem(<[In]> dwFlags As TRANSFER_SOURCE_FLAGS,
                                            <[In]> psiItem As IntPtr,
                                            <[In]> psiDestinationFolder As IntPtr,
                                            <[In], MarshalAs(UnmanagedType.LPWStr)> pszNewName As String,
                                            <[In]> hrCopy As Integer,
                                            <[In]> psiNewlyCreated As IntPtr) As Integer
        <PreserveSig> Function PreDeleteItem(<[In]> dwFlags As TRANSFER_SOURCE_FLAGS,
                                             <[In]> psiItem As IntPtr) As Integer
        <PreserveSig> Function PostDeleteItem(<[In]> dwFlags As TRANSFER_SOURCE_FLAGS,
                                              <[In]> psiItem As IntPtr,
                                              <[In]> hrDelete As Integer,
                                              <[In]> psiNewlyCreated As IntPtr) As Integer
        <PreserveSig> Function PreNewItem(<[In]> dwFlags As TRANSFER_SOURCE_FLAGS,
                                          <[In]> psiDestinationFolder As IntPtr,
                                          <[In], MarshalAs(UnmanagedType.LPWStr)> pszNewName As String) As Integer
        <PreserveSig> Function PostNewItem(<[In]> dwFlags As TRANSFER_SOURCE_FLAGS,
                                           <[In]> psiDestinationFolder As IntPtr,
                                           <[In], MarshalAs(UnmanagedType.LPWStr)> pszNewName As String,
                                           <[In], MarshalAs(UnmanagedType.LPWStr)> pszTemplateName As String,
                                           <[In]> dwFileAttributes As FILE_ATTRIBUTE,
                                           <[In]> hrNew As Integer,
                                           <[In]> psiNewItem As IntPtr) As Integer
        <PreserveSig> Function UpdateProgress(<[In]> iWorkTotal As UInteger,
                                              <[In]> iWorkSoFar As UInteger) As Integer

        ' Not supported!
        <PreserveSig> Function ResetTimer() As Integer
        ' Not supported!
        <PreserveSig> Function PauseTimer() As Integer
        ' Not supported!
        <PreserveSig> Function ResumeTimer() As Integer

    End Interface
#End Region

#Region "Implements IFileOperationProgressSink"
    Private Function StartOperations() As Integer Implements IFileOperationProgressSink.StartOperations
        RaiseEvent OperationsStart()
        Return S_OK
    End Function

    Private Function FinishOperations(<[In]> hrResult As Integer) As Integer Implements IFileOperationProgressSink.FinishOperations
        RaiseEvent OperationsFinish(hrResult)
        Return S_OK
    End Function

    Private Function PreRenameItem(<[In]> dwFlags As TRANSFER_SOURCE_FLAGS,
                                   <[In]> psiItem As IntPtr,
                                   <[In]> <MarshalAs(UnmanagedType.LPWStr)> pszNewName As String) As Integer Implements IFileOperationProgressSink.PreRenameItem
        RaiseEvent OperationsPreRenameItem(dwFlags, psiItem, pszNewName)
        Return S_OK
    End Function

    Private Function PostRenameItem(<[In]> dwFlags As TRANSFER_SOURCE_FLAGS,
                                    <[In]> psiItem As IntPtr,
                                    <[In]> <MarshalAs(UnmanagedType.LPWStr)> pszNewName As String,
                                    <[In]> hrRename As Integer,
                                    <[In]> psiNewlyCreated As IntPtr) As Integer Implements IFileOperationProgressSink.PostRenameItem
        RaiseEvent OperationsPostRenameItem(dwFlags, psiItem, pszNewName, hrRename, psiNewlyCreated)
        Return S_OK
    End Function

    Private Function PreMoveItem(<[In]> dwFlags As TRANSFER_SOURCE_FLAGS,
                                 <[In]> psiItem As IntPtr,
                                 <[In]> psiDestinationFolder As IntPtr,
                                 <[In]> <MarshalAs(UnmanagedType.LPWStr)> pszNewName As String) As Integer Implements IFileOperationProgressSink.PreMoveItem
        RaiseEvent OperationsPreMoveItem(dwFlags, psiItem, psiDestinationFolder, pszNewName)
        Return S_OK
    End Function

    Private Function PostMoveItem(<[In]> dwFlags As TRANSFER_SOURCE_FLAGS,
                                  <[In]> psiItem As IntPtr,
                                  <[In]> psiDestinationFolder As IntPtr,
                                  <[In]> <MarshalAs(UnmanagedType.LPWStr)> pszNewName As String,
                                  <[In]> hrMove As Integer,
                                  <[In]> psiNewlyCreated As IntPtr) As Integer Implements IFileOperationProgressSink.PostMoveItem
        RaiseEvent OperationsPostMoveItem(dwFlags, psiItem, psiDestinationFolder, pszNewName, hrMove, psiNewlyCreated)
        Return S_OK
    End Function

    Private Function PreCopyItem(<[In]> dwFlags As TRANSFER_SOURCE_FLAGS,
                                 <[In]> psiItem As IntPtr,
                                 <[In]> psiDestinationFolder As IntPtr,
                                 <[In]> <MarshalAs(UnmanagedType.LPWStr)> pszNewName As String) As Integer Implements IFileOperationProgressSink.PreCopyItem
        RaiseEvent OperationsPreCopyItem(dwFlags, psiItem, psiDestinationFolder, pszNewName)
        Return S_OK
    End Function

    Private Function PostCopyItem(<[In]> dwFlags As TRANSFER_SOURCE_FLAGS,
                                  <[In]> psiItem As IntPtr,
                                  <[In]> psiDestinationFolder As IntPtr,
                                  <[In]> <MarshalAs(UnmanagedType.LPWStr)> pszNewName As String,
                                  <[In]> hrCopy As Integer,
                                  <[In]> psiNewlyCreated As IntPtr) As Integer Implements IFileOperationProgressSink.PostCopyItem
        RaiseEvent OperationsPostCopyItem(dwFlags, psiItem, psiDestinationFolder, pszNewName, hrCopy, psiNewlyCreated)
        Return S_OK
    End Function

    Private Function PreDeleteItem(<[In]> dwFlags As TRANSFER_SOURCE_FLAGS,
                                   <[In]> psiItem As IntPtr) As Integer Implements IFileOperationProgressSink.PreDeleteItem
        RaiseEvent OperationsPreDeleteItem(dwFlags, psiItem)
        Return S_OK
    End Function

    Private Function PostDeleteItem(<[In]> dwFlags As TRANSFER_SOURCE_FLAGS,
                                    <[In]> psiItem As IntPtr,
                                    <[In]> hrDelete As Integer,
                                    <[In]> psiNewlyCreated As IntPtr) As Integer Implements IFileOperationProgressSink.PostDeleteItem
        RaiseEvent OperationsPostDeleteItem(dwFlags, psiItem, hrDelete, psiNewlyCreated)
        Return S_OK
    End Function

    Private Function PreNewItem(<[In]> dwFlags As TRANSFER_SOURCE_FLAGS,
                                <[In]> psiDestinationFolder As IntPtr,
                                <[In]> <MarshalAs(UnmanagedType.LPWStr)> pszNewName As String) As Integer Implements IFileOperationProgressSink.PreNewItem
        RaiseEvent OperationsPreNewItem(dwFlags, psiDestinationFolder, pszNewName)
        Return S_OK
    End Function

    Private Function PostNewItem(<[In]> dwFlags As TRANSFER_SOURCE_FLAGS,
                                 <[In]> psiDestinationFolder As IntPtr,
                                 <[In]> <MarshalAs(UnmanagedType.LPWStr)> pszNewName As String,
                                 <[In]> <MarshalAs(UnmanagedType.LPWStr)> pszTemplateName As String,
                                 <[In]> dwFileAttributes As FILE_ATTRIBUTE,
                                 <[In]> hrNew As Integer,
                                 <[In]> psiNewItem As IntPtr) As Integer Implements IFileOperationProgressSink.PostNewItem
        RaiseEvent OperationsPostNewItem(dwFlags, psiDestinationFolder, pszNewName, pszTemplateName, dwFileAttributes, hrNew, psiNewItem)
        Return S_OK
    End Function

    Private Function UpdateProgress(<[In]> iWorkTotal As UInteger,
                                    <[In]> iWorkSoFar As UInteger) As Integer Implements IFileOperationProgressSink.UpdateProgress
        RaiseEvent OperationsUpdateProgress(iWorkTotal, iWorkSoFar)
        Return S_OK
    End Function

    Private Function ResetTimer() As Integer Implements IFileOperationProgressSink.ResetTimer
        ' Not supported!
        Return S_OK
    End Function

    Private Function PauseTimer() As Integer Implements IFileOperationProgressSink.PauseTimer
        ' Not supported!
        Return S_OK
    End Function

    Private Function ResumeTimer() As Integer Implements IFileOperationProgressSink.ResumeTimer
        ' Not supported!
        Return S_OK
    End Function

#End Region

End Class
